import math, util
import numpy as np
np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
from scipy.linalg.blas import drot, drotg
import sys; sys.path.append('../')
from models.threestate import Threestate
from models.onestate import Onestate
from innovations import Innovs

def main():
    processor = Modern()
    # model = Onestate()
    model = Threestate()
    # processor.spkf(model)
    processor.spkfcholeksy(model)
    processor.innovs.plot()

class Modern():
    '''modern sigma-point or ukf filter'''
    
    def __init__(self, *args, **kwargs):
        self.args, self.kwargs, self.innovs = args, kwargs, Innovs()

    def spkf(self, model):
        '''sigma-point sampling kalman filter'''
        steps, f, h, F, H, R, x, P = model.ekf()
        f, h, X1, X2, W = model.sp()
        for t, xt, yt in steps():
            X = f(X1(x, P))
            Y = h(X2(X))
            x = W @ X
            y = W @ Y
            P = W @ np.power(X - x, 2) + model.Rww
            K = (W @ np.multiply(X - x, Y - y)) / (W @ np.power(Y - y, 2) + R)
            x = x + K * (yt - y)
            P = P - K * (W @ np.power(Y - y, 2) + R) * K
            self.innovs.add(t, xt, yt, x, y)

    def spkfcholeksy(self, model):
        '''cholesky factorized sigma-point sampling kalman filter'''
        steps, f, h, F, H, R, x, P = model.ekf()
        f, h, X1, X2, W, S = model.sp()
        for t, xt, yt in steps():
            x, S, X = self.temporal(x, S, model)
            x, S, y = self.observational(x, S, X, yt, model)
            self.innovs.add(t, xt, yt, x, y)

    def cholupdate(self, R, z):
        '''cholesky update'''
        n = z.shape[0]
        for k in range(n):
            c, s = drotg(R[k, k], z[k])
            drot(R[k, :], z, c, s, overwrite_x=True, overwrite_y=True)
        return R

    def choldowndate(self, R, z):
        '''cholesky downdate'''
        n = R.shape[0]
        for k in range(n):
            if (R[k, k] - z[k]) * (R[k, k] + z[k]) < 0: return R
            rbar = np.sqrt((R[k, k] - z[k]) * (R[k, k] + z[k]))
            for j in range(k + 1, n):
                R[k, j] = 1. / rbar * (R[k, k] * R[k, j] - z[k] * z[j])
                z[j] = 1. / R[k, k] * (rbar * z[j] - z[k] * R[k, j])
            R[k, k] = rbar
        return R

    def temporal(self, xhat, S, model):
        '''cholesky factorized temporal update'''
        X = model.SPKF.vf(model.SPKF.X1(xhat, S))
        xhat = model.SPKF.Wm @ X.T
        for i in range(7): model.SPKF.Xtil[:, i] = X[:, i] - xhat
        q, r = np.linalg.qr(np.concatenate([math.sqrt(model.SPKF.Wc[1]) * model.SPKF.Xtil[:, 1:], model.SPKF.Sw], 1))
        S = self.cholupdate(r.T[0:3, 0:3], model.SPKF.Wc[0] * model.SPKF.Xtil[:, 0])
        return xhat, S, X

    def observational(self, xhat, S, X, obs, model):
        '''cholesky factorized observational update'''
        Y = model.SPKF.vh(model.SPKF.X2(X))
        yhat = model.SPKF.Wm @ Y.T
        for i in range(7): model.SPKF.Ytil[0, i] = Y[i] - yhat
        q, r = np.linalg.qr(np.concatenate([math.sqrt(model.SPKF.Wc[1]) * model.SPKF.Ytil[:, 1:], model.SPKF.Sv], 1))
        Sy = self.cholupdate(r.T[0:1, 0:1], model.SPKF.Wc[0] * model.SPKF.Ytil[:, 0])
        for i in range(7): model.SPKF.Pxy[:, 0] = model.SPKF.Pxy[:, 0] + model.SPKF.Wc[i] * model.SPKF.Xtil[:, i] * model.SPKF.Ytil.T[i, :]
        if Sy[0, 0] < math.sqrt(10) or Sy[0, 0] > math.sqrt(1000): Sy[0, 0] = math.sqrt(1000)
        K = model.SPKF.Pxy / Sy[0, 0] ** 2
        U = K * Sy
        xhat = xhat + K[:, 0] * (obs - yhat)
        S = self.choldowndate(S, U)
        return xhat, S, yhat

if __name__ == "__main__":
    main()




import math
import numpy as np
from modelbase import ModelBase

class Onestate(ModelBase):
    '''one-state reference model.'''

    def ekf(self): return self.steps, self.f, self.h, self.F, self.H, self.R, self.x0, self.P0
    def ekfud(self): return self.steps, self.f, self.h, self.F, self.H, self.R, self.x0, self.P0, self.G, self.Q
    def sp(self): return self.SPKF.vf, self.SPKF.vh, self.SPKF.X1, self.SPKF.X2, self.SPKF.W
    def pf(self): return self.PF.nsamp, self.PF.F, self.PF.H

    def init(self):
        self.tsteps = 151
        self.dt = .01
        self.x = np.array([2.])
        self.Rww = 1e-6 * np.array([1])
        self.R = 9e-2
        self.x0 = np.array([2.2])
        self.P0 = .01 * np.eye(1)
        self.G = np.eye(1)
        self.Q = self.Rww * np.eye(1)
        self.SPKF = SPKF(self)
        self.PF = PF(self)

    def steps(self):
        for tstep in range(self.tsteps):
            tsec = tstep * self.dt
            self.x = self.f(self.x)
            self.y = self.h(self.x)
            if tstep == 0: continue
            self.log.append([tsec, self.x, self.y])
            yield (tsec, self.x, self.y)

    def f(self, x):
        w = math.sqrt(self.Rww) * np.random.randn()
        return (1 - .05 * self.dt) * x + (.04 * self.dt) * x ** 2 + w

    def h(self, x):
        v = math.sqrt(self.R) * np.random.randn()
        return x ** 2 + x ** 3 + v

    def F(self, x):
        A = np.eye(1)
        A[0, 0] = 1 - .05 * self.dt + .08 * self.dt * x
        return A

    def H(self, x):
        return 2 * x + 3 * x ** 2

class SPKF():
    def __init__(self, parent):
        self.parent = parent
        self.vf = np.vectorize(parent.f)
        self.vh = np.vectorize(parent.h)
        bignsubx = 1
        kappa = 1
        k0 = bignsubx + kappa
        k1 = kappa / float(k0)
        k2 = 1 / float(2 * k0)
        self.W = np.array([k1, k2, k2])
        self.k0 = k0
        self.kappa = kappa

    def X1(self, xhat, Ptil):
        return [xhat, xhat + math.sqrt(self.k0 * Ptil), xhat - math.sqrt(self.k0 * Ptil)]

    def X2(self, X):
        return [X[0], X[1] + self.kappa * math.sqrt(self.parent.Rww), X[2] - self.kappa * math.sqrt(self.parent.Rww)]

class PF():
    def __init__(self, parent):
        self.parent = parent
        self.xhat0 = 2.05
        self.nsamp = 250

    def F(self, x):
        return (1 - .05 * self.parent.dt) * x + (.04 * self.parent.dt) * x ** 2 + math.sqrt(self.parent.Rww) * np.random.randn(self.nsamp)

    def H(self, y, x):
        return np.exp(-np.log(2. * np.pi * self.parent.R) / 2. - (y - x ** 2 - x ** 3) ** 2 / (2. * self.parent.R))

if __name__ == "__main__":
    pass




import math
import numpy as np
from modelbase import ModelBase

class BearingsOnly(ModelBase):
    '''basically about being on a ww2 era sub. your sub is travelling along, and you begin hearing the sound of a ship at some bearing. over time, as the bearing changes, you can estimate the position and velocity of the ship. both your sub and the ship are moving the entire time. at some point you make a course change for your sub to pursue the ship.'''

    pass


import math
import numpy as np
import matplotlib.pyplot as plt

class Innovs():
    def __init__(self):
        self.log = []

    def add(self, t, xt, yt, x, y):
        try: self.log.append([t, x[0], y, xt[0] - x[0], yt - y])
        except: self.log.append([t, x, y, xt[0] - x, yt - y])

    def update2(self, t, xhat, yhat, err, inn, Ree, Ptil):
        self.log.append([t, xhat, yhat, err, inn, Ree, Ptil])

    def autocorr1(self, x):
        n = x.size
        ac = np.zeros([n, 1])
        for k in range(1, n):  # k = 1 to n-1
            ac[k] = 0
            for t in range(1, n - k):  # t = 1 to n-k
                ac[k] += (self.ets[t - 1] - self.mhate) * (self.ets[t - 1 + k] - self.mhate)  # / (n-k)
        return ac

    def autocorr2(self, x):
        ac = np.correlate(x, x, mode='full')
        return ac[ac.size // 2:]

    def finalize(self):
        self.log = np.asarray(self.log)
        self.tts = self.log[:, 0]
        self.ets = self.log[:, 4]
        self.mhate = np.mean(self.ets)
        self.Rhatee = np.mean(np.power(self.ets, 2))
        n = len(self.ets)
        self.tau = 1.96 * math.sqrt(self.Rhatee / n)
        self.iszeromean = True
        if abs(self.mhate) > self.tau:
            self.iszeromean = False
        self.zmw = self.autocorr1(self.ets)

    def plot(self):
        self.finalize()
        lw = 1
        plt.subplot(3, 2, 1), plt.plot(self.tts, self.log[:, 1], linewidth=lw), plt.ylabel('xhat')
        plt.subplot(3, 2, 2), plt.plot(self.tts, self.log[:, 3], linewidth=lw), plt.ylabel('err')
        plt.subplot(3, 2, 3), plt.plot(self.tts, self.log[:, 2], linewidth=lw), plt.ylabel('yhat')
        plt.subplot(3, 2, 4), plt.plot(self.tts, self.log[:, 4], linewidth=lw), plt.ylabel('innov')
        plt.subplot(3, 2, 5), plt.plot(self.tts, self.zmw, linewidth=lw), plt.ylabel('autocorr')
        plt.show()


import numpy as np
np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
import sys; sys.path.append('../'); sys.path.append('../cmake-build-debug/libstatespace')
from models.threestate import Threestate
from models.onestate import Onestate
from innovations import Innovs
import libstatespace
api = libstatespace.Api()

def main():
    processor = Classical()
    model = Onestate()
    # model = Threestate()
    processor.ekf(model)
    # processor.ekfud(model)
    processor.innovs.plot()

class Classical():
    '''classical kalman filter'''

    def __init__(self, *args, **kwargs):
        self.args, self.kwargs, self.innovs = args, kwargs, Innovs()

    def ekf(self, model):
        '''basic form'''
        steps, f, h, F, H, R, x, P = model.ekf()
        for t, xt, yt in steps():
            x = f(x)
            y = h(x)
            P = F(x) @ P @ F(x)
            K = P @ H(x) / (H(x) @ P @ H(x) + R)
            x = x + K * (yt - y)
            P = (1 - K * H(x)) * P
            self.innovs.add(t, xt, yt, x, y)

    def ekfud(self, model):
        '''UD factorized form'''
        steps, f, h, F, H, R, x, P, G, Q = model.ekfud()
        U, D = self.udfactorize(P)
        for t, xt, yt in steps():
            x, U, D = self.temporal(f(x), U, D, F(x), G, Q)
            x, U, D, y = self.observational(x, U, D, H(x), yt, R, h(x))
            self.innovs.add(t, xt, yt, x, y)

    def ekfudcpp(self, model):
        '''UD factorized form in cpp'''
        steps, f, h, F, H, R, x, P, G, Q = model.ekfud()
        ud = api.udfactorize(P); U, D = ud[0], np.diag(ud[1].transpose()[0])
        for t, xt, yt in steps():
            cpp = api.temporal(f(x), U, D, H(x), G, Q)
            x, U, D = cpp[0].flatten(), cpp[1], cpp[2]
            cpp = api.observational(x, U, D, H(x), yt, R, h(x))
            x, U, D, y = cpp[0].flatten(), cpp[1], cpp[2], h(x)
            self.innovs.add(t, xt, yt, x, y)

    def udfactorize(self, M):
        '''UD factorization'''
        assert np.allclose(M, M.T)
        n, M = M.shape[0], np.triu(M)
        U, d = np.eye(n), np.zeros(n)
        for j in reversed(range(2, n + 1)):
            d[j - 1] = M[j - 1, j - 1]
            if d[j - 1] > 0: alpha = 1.0 / d[j - 1]
            else: alpha = 0.0
            for k in range(1, j):
                beta = M[k - 1, j - 1]
                U[k - 1, j - 1] = alpha * beta
                M[0:k, k - 1] = M[0:k, k - 1] - beta * U[0:k, j - 1]
        d[0] = M[0, 0]
        return U, np.diag(d)

    def temporal(self, xin, Uin, Din, Phi, Gin, Q):
        '''thornton temporal update'''
        U, D, G, n, r = np.eye(len(xin)), Din, Gin, len(xin), len(xin)
        x, PhiU = Phi @ xin, Phi @ Uin
        for i in reversed(range(len(xin))):
            sigma = 0
            for j in range(n):
                sigma = sigma + PhiU[i, j] ** 2 * Din[j, j]
                if (j <= r - 1): sigma = sigma + G[i, j] ** 2 + Q[j, j]
            D[i, i] = sigma
            ilim = i - 1
            if ilim > 0:
                for j in range(ilim):
                    sigma = 0
                    for k in range(n): sigma = sigma + PhiU[i, k] * Din[k, k] * PhiU[j, k]
                    for k in range(r): sigma = sigma + G[i, k] * Q[k, k] * G[j, k]
                    U[j, i] = sigma / D[i, i]
                    for k in range(n): PhiU[j, k] = PhiU[j, k] - U[j, i] * PhiU[i, k]
                    for k in range(r): G[j, k] = G[j, k] - U[j, i] * G[i, k]
        return x, U, D

    def observational(self, xin, Uin, Din, H, obs, R, yhat):
        '''bierman observation update'''
        x, U, D, dz, alpha, gamma = xin, Uin, Din, obs - yhat, R, 1/R
        a = U.T @ H.T
        b = D @ a
        for j in range(len(xin)):
            beta = alpha
            alpha = alpha + a[j] * b[j]
            lamda = -a[j] * gamma
            gamma = 1 / alpha
            D[j, j] = beta * gamma * D[j, j]
            jlim = j - 1
            if not jlim < 0:
                for i in range(jlim):
                    beta = U[i, j]
                    U[i, j] = beta + b[i] * lamda
                    b[i] = b[i] + b[j] * beta
        dzs = gamma * dz
        x = x + dzs * b
        return x, U, D, yhat

if __name__ == "__main__":
    main()




import math
import numpy as np
from modelbase import ModelBase

class Threestate(ModelBase):
    '''three-state reference model.'''

    def ekf(self): return self.steps, self.f, self.h, self.F, self.H, self.R, self.x0, self.P0
    def ekfud(self): return self.steps, self.f, self.h, self.F, self.H, self.R, self.x0, self.P0, self.G, self.Q
    def sp(self): return self.SPKF.vf, self.SPKF.vh, self.SPKF.X1, self.SPKF.X2, self.SPKF.W, self.SPKF.S
    def pf(self): return self.PF.nsamp, self.PF.F, self.PF.H

    def init(self):
        self.tsteps = 1501
        self.dt = .01
        self.x = np.array([2., .05, .04])
        self.Rww = 1e-9 * np.array([1, 1, 1])
        self.R = 9e-2
        self.x0 = np.array([2, .055, .044])
        self.P0 = 1. * np.eye(3)
        self.G = np.eye(3)
        self.Q = np.diag(self.Rww)
        self.SPKF = SPKF(self)
        self.PF = PF(self)

    def steps(self):
        for tstep in range(self.tsteps):
            tsec = tstep * self.dt
            self.x = self.f(self.x)
            self.y = self.h(self.x)
            if tstep == 0: continue
            self.log.append([tsec, self.x, self.y])
            yield (tsec, self.x, self.y)

    def f(self, x):
        w = np.multiply(np.random.randn(1, 3), np.sqrt(np.diag(self.Rww)))
        return np.array([(1 - x[1] * self.dt) * x[0] + x[2] * self.dt * x[0] ** 2, x[1], x[2]]) + np.diag(w)

    def h(self, x):
        v = math.sqrt(self.R) * np.random.randn()
        return x[0] ** 2 + x[0] ** 3 + v

    def F(self, x):
        A = np.eye(3)
        A[0, 0] = 1 - x[1] * self.dt + 2 * x[2] * self.dt * x[0]
        A[0, 1] = -self.dt * x[0]
        A[0, 2] = self.dt * x[0] ** 2
        return A

    def H(self, x):
        return np.array([2 * x[0] + 3 * x[0] ** 2, 0, 0])

class SPKF():
    def __init__(self, parent):
        self.parent = parent
        P0 = .1 * np.eye(3)
        self.S = np.linalg.cholesky(P0)
        self.Sw = np.linalg.cholesky(np.diag(parent.Rww))
        self.Sv = np.linalg.cholesky(np.diag(parent.R * np.array([1])))
        n = 3
        kappa = 1
        alpha = 1
        beta = 2
        lam = alpha ** 2 * (n + kappa) - n
        wi = 1 / float(2 * (n + lam))
        w0m = lam / float(n + lam)
        w0c = lam / float(n + lam) + (1 - alpha ** 2 + beta)
        self.Wm = np.array([w0m, wi, wi, wi, wi, wi, wi])
        self.Wc = np.array([w0c, wi, wi, wi, wi, wi, wi])
        self.nlroot = math.sqrt(n + lam)
        self.Xtil = np.zeros((3, 7))
        self.Ytil = np.zeros((1, 7))
        self.Pxy = np.zeros((3, 1))
        self.W = self.Wm

    def X1(self, x, C):
        X = np.zeros([3, 7])
        X[:, 0] = x
        X[:, 1] = x + self.nlroot * C.T[:, 0]
        X[:, 2] = x + self.nlroot * C.T[:, 1]
        X[:, 3] = x + self.nlroot * C.T[:, 2]
        X[:, 4] = x - self.nlroot * C.T[:, 0]
        X[:, 5] = x - self.nlroot * C.T[:, 1]
        X[:, 6] = x - self.nlroot * C.T[:, 2]
        return X

    def X2(self, X):
        Xhat = np.zeros([3, 7])
        Xhat[:, 0] = X[:, 0]
        Xhat[:, 1] = X[:, 1] + self.nlroot * self.Sw.T[:, 0]
        Xhat[:, 2] = X[:, 2] + self.nlroot * self.Sw.T[:, 1]
        Xhat[:, 3] = X[:, 3] + self.nlroot * self.Sw.T[:, 2]
        Xhat[:, 4] = X[:, 4] - self.nlroot * self.Sw.T[:, 0]
        Xhat[:, 5] = X[:, 5] - self.nlroot * self.Sw.T[:, 1]
        Xhat[:, 6] = X[:, 6] - self.nlroot * self.Sw.T[:, 2]
        return Xhat

    def vf(self, X):
        for i in range(7): X[:, i] = self.parent.f(X[:, i])
        return X

    def vh(self, Xhat):
        Y = np.zeros(7)
        for i in range(7): Y[i] = self.parent.h(Xhat[:, i])
        return Y

    def Xtil(self, X, W):
        Xtil = np.zeros((3, 7))
        for i in range(7): Xtil[:, i] = X[:, i] - W @ X.T
        return Xtil

    def ksi(self, Y, W):
        ksi = np.zeros((1, 7))
        for i in range(7): ksi[0, i] = Y[i] - W @ Y.T
        return ksi

class PF():
    def __init__(self, parent):
        self.parent = parent
        self.xhat0 = np.array([2.0, .055, .044])
        self.nsamp = 250

    def F(self, x):
        return (1 - x[1] * self.parent.dt) * x[0] + x[2] * self.parent.dt * x[0] ** 2

    def H(self, y, x):
        return np.exp(-np.log(2. * np.pi * self.parent.R) / 2. - (y - x ** 2 - x ** 3) ** 2 / (2. * self.parent.R))

if __name__ == "__main__":
    pass




import numpy as np
np.warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)
import math, util
import sys; sys.path.append('../')
from models.threestate import Threestate
from models.onestate import Onestate
from innovations import Innovs

def main():
    processor = Particle()
    # model = Onestate()
    model = Threestate()
    # processor.pf1(model)
    processor.pf2(model)
    processor.innovs.plot()

class Particle():
    '''particle filter, sequential monte carlo sampling processor, bootstrap particle filter'''
    
    def __init__(self, *args, **kwargs):
        self.args, self.kwargs, self.innovs = args, kwargs, Innovs()

    def pf1(self, model): # todo still scalar
        '''particle filter'''
        steps, f, h, F, H, R, x, P = model.ekf()
        nsamp, F, H = model.pf()
        xp = x + math.sqrt(1e-2) * np.random.randn(nsamp)
        W = self.normalize(np.ones(nsamp))
        for t, xt, yt in steps():
            xp = self.resample(xp, W)
            xp = F(xp)
            y = h(x)
            W = self.normalize(H(yt, xp))
            x = W @ xp
            self.innovs.add(t, xt, yt, x, y)

    def pf2(self, model):
        '''particle filter'''
        steps, f, h, F, H, R, x, P = model.ekf()
        nsamp, F, H = model.pf()
        xp = x + np.sqrt(model.Rww) * np.random.randn(model.PF.nsamp, len(x))
        W = np.ones((nsamp, 3)) / nsamp
        for t, xt, yt in steps():
            xp[:, 0], xp[:, 1], xp[:, 2] = self.resample(xp[:, 0], W[:, 0]), self.roughen(xp[:, 1]), self.roughen(xp[:, 2])
            xp[:, 0] = np.apply_along_axis(F, 1, xp)
            y = model.h(x)
            W[:, 0] = self.normalize(H(yt, xp[:, 0]))
            x = [W[:, 0].T @ xp[:, 0], W[:, 1].T @ xp[:, 1], W[:, 2].T @ xp[:, 2]]
            self.innovs.add(t, xt, yt, x, y)

    def resample(self, xi, Wi):
        '''particle resampling.'''
        tmp = []
        for i in range(xi.size):
            tmp.append([xi[i], Wi[i]])
        tmp = sorted(tmp, key=lambda x: x[0])
        cdf = [[tmp[0][0], tmp[0][1]]]
        for i in range(1, len(tmp)):
            cdf.append([tmp[i][0], tmp[i][1] + cdf[i - 1][1]])
        cdf = np.asarray(cdf)
        uk = np.sort(np.random.uniform(size=xi.size))
        xhati, k = [], 0
        for row in cdf:
            while k < uk.size and uk[k] <= row[1]:
                xhati.append(row[0])
                k += 1
        xhati = np.asarray(xhati)
        assert xhati.size == xi.size
        return xhati

    def roughen(self, x):
        return x + .1 * np.random.randn(x.size)

    def normalize(self, W):
        return W / sum(W)

if __name__ == "__main__":
    main()
